{{
/**
 * Peter Vikaik's heap object inspires this more flexible malloc.
 * This is a malloc object. The malloc area is a large contiguous byte array.
 *
 * Features:
 *
 *      Small footprint
 *      Data type size (word or long) is selectable at startup.
 *      Allocation of blocks is done using a first-fit scheme.
 *      Malloc area can be increased if necessary and memory is available.
 *      Freed blocks are reusable.
 *      Contiguous freed blocks are concatenated when possible.
 */

/*
  Implementation
  heapStart:   DW   NEXT1        //pointer to next block NEXT1  := (word[heapStart] & $FFFE) (b0 set if this block allocated)
                                 //number of bytes in the block := (word[heapStart] & $FFFE)-heapStart
  <blkaddr>    DS   SIZE1        //blkaddr is alloc'd address   := NEXT1 + 2
  NEXT1:       DW   NEXT2        // NEXT2 := (word[NEXT1] & $FFFE)
                                 //number of bytes in the block := (word[NEXT1] & $FFFE)-NEXT1
  <blkaddr>    DS   SIZE2        //blkaddr is alloc'd address   := NEXT2 + 2
               ..
               DW   heapEnd
               DW   SIZEn
  <blkaddr>    DS   SIZEn
  heapEnd:     DW   0            //closing 0 word

  Zero entry heap size:  1*sizeof(unit) + 0              | unit = 2, 4, or 8 bytes                 
  One entry heap size :  1*sizeof(unit) + 2*B[1] div 2   | B[1] = 1 byte
  General heap size   :  N*sizeof(unit) + 2*B[N] div 2   | B[N] = num bytes for N'th entry
   4 bytes (word NEXT1, word SIZE1,allocatable byte, word 0)

  Address 7EF0-1 7EF2-3 7EF4-5 7EF6-7 7EF8-9 7EFA-B 7EFC-D 7EFE-F 7F00
  Data     7EF8   0000   0000   0000   7EFC   0000   7EFE   7F00  0000
*/
}}

DAT
  heapStart   long 0    '//address of heap
  heapEnd     long 0    '//points to closing 0 word
  maxHeapSize long 0    '// maximum heap size
  freecnt     long 0    '//track freecounts and scrub if freecnt // FREESCRUB
  freescrub   word 0

{{
/**
 * sizeof_int is the number of bytes in a basic pointer element equivalent to sizeof(int)
 * This is done for type compatability.
 * User sets this value at startup @see init
 */
}}
  sizeof_int  word 0

PUB init(startaddr, size, sint) | endaddr ' intialize with an array or open memory
{{
/**
 * Create a heap from endAddress growing down
 * @param endAddr is end address of heap; this is fixed
 * @param size is initial size of heap - this can be adjusted later.
 * @param sint is the size of an integer or basic data element; this is fixed.
 * @see setHeapSize();
 */
 _ void create(int endAddress);
}}
  sizeof_int := sint
  freescrub := 4                                        ' set default scrub ratio 4 frees/scrub
  freecnt := 0                                          ' initialize free counter for scrub ratio
  endaddr   := (startaddr + size) & !7                  ' make sure we have good end address alignment
  maxHeapSize := endaddr-startaddr                      ' starting heap size - can be changed later
  heapEnd   := endaddr
  heapStart := heapEnd-sizeof_int
  setvalue(heapStart,heapEnd)
  setvalue(heapEnd,0)

PUB malloc(size) | newstart
{{
/**
 * first fit malloc a pointer of size from heap.
 * @param size is size of data to allocate. smallest size is 2 bytes
 * @returns 0 on failure or pointer
 */
 _ void* malloc(int size);
}}
  size := size+sizeof_int
  result := findFree(size)
  if result
    setUsed(result)
    result := block(result)
  else
    result := heapStart-size + size // 2
    if (heapEnd-result) > maxHeapSize
      result := 0
    else
      newstart := result
      setvalue(newstart,heapStart)
      heapStart := newstart
      setUsed(heapStart)
      'print(string("malloc'd"), heapStart, nextptr(heapStart))
      result := block(heapStart)
      'print(string("malloc'd"), heapStart, result)

PUB free(ptr)
{{
/**
 * free a pointer in the heap
 * @param ptr is the block pointer previously returned from malloc
 * @returns 0 on failure or pointer ptr
 */
 _ void* free(void* ptr);
}}
  result := heapPtr(ptr)
  'print(string("free"), result,nextptr(result))
  if isptr(result)
    'print(string("in free"), result, getValue(result))
    setFree(result)
    'isFree(result)
    if freecnt++ // FREESCRUB
      scrub
    'print(string("in free"), result, getValue(result))
    result := ptr  
  else
    result := 0

PRI findFree(size) | ptr
{
/**
 * find a free pointer of size
 * @param size is required size for free entry
 * @returns 0 on failure or pointer
 */ 
 _ void* findFree(int size);
}
  ptr := heapStart
  repeat while(isNotEnd(ptr))
    if isFree(ptr)
      'print(string("malloc got free"), ptr, ptr)
      if (nextptr(ptr)-ptr) > size
        print(string("malloc free sz"), heapStart, nextptr(ptr)-ptr)
        return ptr
    ptr := nextptr(ptr)

PUB getHeapEnd
{{
/**
 * get current heap end.
 * @returns current heap end.
 */
 _ int getHeapEnd();
}}
  result := heapEnd

PUB getHeapSize
{{
/**
 * get heap size.
 * @returns current heap size.
 */
 _ int getHeapSize();
}}
  result := maxHeapSize

PUB getHeapStart
{{
/**
 * get current heap start.
 * @returns current heap start.
 */
 _ int getHeapStart();
}}
  result := heapStart

PUB setHeapSize(size)
{{
/**
 * set heap size.
 * @note this allows increasing size only.
 * @param size is new size of heap
 * @returns non-zero resize failed.
 */
 _ int setHeapSize(int size);
}}
  if size < maxHeapSize 
    return 1
  maxHeapSize := size
  result := 0

PUB setHeapStart(addr)
{{
/**
 * set address where heap should stop growing down
 * @param addr is smallest start address of heap
 * @returns non-zero if update failed. this means there is a problem.
 */
 _ int setHeapStart(int addr);
}}
  if heapStart < addr
    return 1
  maxHeapSize := heapEnd-addr
  result := 0

PUB getSizeOfInt
{{
/**
 * @returns size of an integer in bytes.
 */
 _ int getSizeofInt();
}}
  return sizeof_int
  
PUB setScrubCount(num)
{{
/**
 * Set the freescrub count which is the number of free required for scrub to happen.
 * This count affects performance of free. A bigger number means less scrubbing and
 * more memory fragmentation before scrubbing. A smaller number means more scrubbing.
 * @param addr is smallest start address of heap
 * @returns non-zero if update failed. this means there is a problem.
 */
 _ void setScrubCount(int num);
}}
  if num > 1
    freescrub := num

PUB getValue(ptr)
{{
/**
 * This function gets the default size content of address ptr.
 * @note This is wrapper uses a sizeof_int constant defined above.
 * @param ptr is address of value to fetch
 * @returns value at ptr address.
 */
 _ int getData(void* ptr);
}}
  case getSizeOfInt
    2: result := word[ptr]
    4: result := long[ptr]

PUB setValue(ptr, val)
{{
/**
 * This function sets the default size content of address ptr with value.
 * @note This is wrapper uses a sizeof_int constant defined above.
 * @param ptr is address of value to set
 * @param val is value to set at ptr address
 */
 _ int setValue(void* ptr, int val);
}}
  case getSizeOfInt
    2: word[ptr] := val
    4: long[ptr] := val
    
PUB getHeapPtr(userptr)
{{
/**
 * get the internal heap pointer address of the user ptr.
 * @param userptr ptr returned by malloc.
 * @returns address of the internal heap pointer associated with the user pointer.
 */
 _ int getHeapPtr(void* userptr);
}}
  result := heapptr(userptr)

PUB getUserPtr(heap_ptr)
{{
/**
 * get the user heap pointer address of the heap ptr.
 * @param userptr ptr returned by malloc.
 * @returns address of the user pointer associated with the heap pointer.
 */
 _ int getUserPtr(void* heap_ptr);
}}
  result := block(heap_ptr)

PUB getNextPtr(heap_ptr)
{{
/**
 * get the heap address of the ptr following heapptr.
 * @note heapptr is not a pointer returned by malloc; it is the internal accounting heap pointer.
 * @param heapptr ptr returned by getHeapPtr(userptr)
 * @returns address of the next user pointer.
 */
 _ int getNextPtr(void* heap_ptr);
}}
  result := getValue(heap_ptr) & !1

PUB getNextUserPtr(user_ptr)
{{
/**
 * get the address of the ptr following this block.
 * @param user_ptr is a pointer returned by malloc.
 * @returns address of the next user pointer.
 */
 _ int getNextUserPtr(void* heap_ptr);
}}
  result := getValue(heapPtr(user_ptr)) & !1

PUB nextptr(ptr)
{{ get next heap pointer address }}
  result := getValue(ptr) & !1

PRI setUsed(ptr)
  result := getValue(ptr) | 1
  setValue(ptr,result)

PUB isBlockFree(ptr)
{{
/**
 * find out if user pointer is free.
 * @returns non-zero if block is free.
 */
 _ int isBlockFree(void* ptr);
}}
  result := block(ptr)
  result := getValue(result) & 1
  result := (!result & 1)

PUB isFree(ptr)
{{ pointer is free if bit 0 of getValue(ptr) is clear }}
  result := getValue(ptr) & 1
  result := (!result & 1)
  
PRI setFree(ptr)
  result := getValue(ptr) & !1
  setValue(ptr,result)

PUB block(ptr)
{{ returns the user memory associated with a heap ptr }}
  result := ptr + sizeof_int

PUB heapPtr(ptr)
{{ returns heap accounting associated with the block ptr }}
  result := ptr - sizeof_int

PUB isPtr(ptr) | p1
{{ returns true if ptr is a heap ptr }}
  result := 0
  p1 := heapStart
  repeat while(p1 < heapEnd)
    if p1 == ptr
      return 1
    p1 := nextptr(p1)

PRI linkFree(p1, p2) | ptr
  if isFree(p1) == 0
    return
  ptr := block(p1)
  setValue(p1,nextptr(p2))
  repeat while(ptr < p2) 
    setValue(ptr,$F4EE)
    ptr+=sizeof_int
  
PRI scrub | ptr, p1
{ memory scrubber used to concatenate free memory bits }
  ptr := nextptr(heapStart)
  repeat while(isNotEnd(ptr))
    'printentry(ptr,".")
    'printentry(nextptr(ptr),".")
    'print(string("scrub1"), isFree(ptr), ptr)
    'print(string("scrub2"), isFree(nextptr(ptr)),nextptr(ptr))
    if isFree(ptr) and isFree(nextptr(ptr))
      p1 := ptr
      repeat while(isFree(nextptr(ptr)) and (nextptr(nextptr(ptr)) <> 0))
        ptr := nextptr(ptr)
        'print(string("scrub"), p1, ptr)
      linkFree(p1,ptr)
      next ' let main loop handle advancing pointer
    ptr := nextptr(ptr)

PUB isNotEnd(ptr)
{{ returns true if ptr is not at end of list }}
  result := nextptr(ptr) < heapEnd and (nextptr(ptr) <> 0)

'PUB isEnd(ptr) ' interesting but unused
'  result := nextptr(ptr) => heapEnd and (nextptr(ptr) == 0)

PRI print(fn, ptr, val)
{
/**
 * print string and two variables in hex
 */
 _ void print(char* fn, int ptr, int val);
}
{
  vga.out($d)
  vga.str(fn)
  vga.out($9)
  vga.hex(ptr, (getSizeofInt<<1))
  vga.out(":")
  vga.hex(val, (getSizeofInt<<1))
OBJ vga: "VGA_Text"
}
